"""Functions for parsing the parameter data files"""

import yaml
import pkgutil
from flavio.classes import *
from flavio.statistics.probability import *
from flavio._parse_errors import errors_from_string
import flavio
import re
from flavio.measurements import _fix_correlation_matrix
from math import sqrt
from particle import Particle, data as p_data


def _read_yaml_object_metadata(obj, constraints):
    parameters = yaml.safe_load(obj)
    for parameter_name, info in parameters.items():
        p = Parameter(parameter_name)
        if 'description' in info and info['description'] is not None:
            p.description = info['description']
        if 'tex' in info and info['tex'] is not None:
            p.tex = info['tex']

def read_file_metadata(filename, constraints):
    """Read parameter values from a YAML file."""
    with open(filename, 'r') as f:
        _read_yaml_object_metadata(f, constraints)

def _read_yaml_object_values(obj, constraints):
    parameters = yaml.safe_load(obj)
    for parameter_name, value in parameters.items():
        p = Parameter[parameter_name] # this will raise an error if the parameter doesn't exist!
        constraints.set_constraint(parameter_name, value)

def _read_yaml_object_new(obj):
    """Read parameter constraints from a YAML stream or file that are compatible
    with the format generated by the `get_yaml` method of
    `flavio.classes.ParameterConstraints`."""
    parameters = yaml.safe_load(obj)
    return ParameterConstraints.from_yaml_dict(parameters)

def _read_yaml_object_values_correlated(obj, constraints):
    list_ = yaml.safe_load(obj)
    for parameter_group in list_:
        parameter_names = []
        central_values = []
        errors = []
        for dict_list in parameter_group['values']:
            parameter_name, value = list(dict_list.items())[0]
            Parameter[parameter_name] # this will raise an error if the parameter doesn't exist!
            parameter_names.append(parameter_name)
            error_dict = errors_from_string(value)
            central_values.append(error_dict['central_value'])
            squared_error = 0.
            for sym_err in error_dict['symmetric_errors']:
                squared_error += sym_err**2
            for asym_err in error_dict['asymmetric_errors']:
                squared_error += asym_err[0]*asym_err[1]
            errors.append(sqrt(squared_error))
        correlation = _fix_correlation_matrix(parameter_group['correlation'], len(parameter_names))
        covariance = np.outer(np.asarray(errors), np.asarray(errors))*correlation
        if not np.all(np.linalg.eigvals(covariance) > 0):
            # if the covariance matrix is not positive definite, try a dirty trick:
            # multiply all the correlations by 0.99.
            n_dim = len(correlation)
            correlation = (correlation - np.eye(n_dim))*0.99 + np.eye(n_dim)
            covariance = np.outer(np.asarray(errors), np.asarray(errors))*correlation
            # if it still isn't positive definite, give up.
            assert np.all(np.linalg.eigvals(covariance) > 0), "The covariance matrix is not positive definite!" + str(covariance)
        constraints.add_constraint(parameter_names, MultivariateNormalDistribution(central_values, covariance))

def read_file(filename):
    """Read parameter values from a YAML file in the format generated by the
    `get_yaml` method of the `ParameterConstraints` class, returning a
    `ParameterConstraints` instance."""
    with open(filename, 'r') as f:
        return _read_yaml_object_new(f)

def read_file_values(filename, constraints):
    """Read parameter values from a YAML file."""
    with open(filename, 'r') as f:
        _read_yaml_object_values(f, constraints)

def read_file_values_correlated(filename, constraints):
    """Read parameter values from a YAML file."""
    with open(filename, 'r') as f:
        _read_yaml_object_values_correlated(f, constraints)

def write_file(filename, constraints):
    """Write parameter constraints to a YAML file."""
    with open(filename, 'w') as f:
        yaml.dump(constraints.get_yaml_dict(), f)

class FlavioParticle(Particle):
    """This class extends the `particle.Particle` class.

    Additional class methods
    ------------------------
    - from_flavio_name(flavio_name)
      returns a class instance for a given `flavio_name`
    - flavio_all()
      returns a set of all class instances used in flavio

    Additional properties
    ---------------------
    - flavio_name
      the particle name as used in flavio if defined, otherwise `None`
    - latex_name_simplified
      a simplified version of the latex name returned by `latex_name`
    - flavio_m
      a tuple with data on the particle mass as used in flavio, containing
      entries `name`, `tex`, `description`, `central`, `right`, `left`
    - flavio_tau
      a tuple with data on the particle lifetime as used in flavio, containing
      entries `name`, `tex`, `description`, `central`, `right`, `left`
    """

    PDG_PARTICLES = {
        'Bs': 531,
        'Bc': 541,
        'Bs*': 533,
        'B*+': 523,
        'B*0': 513,
        'B+': 521,
        'B0': 511,
        'Ds': 431,
        'Ds*': 433,
        'D+': 411,
        'D0': 421,
        'h': 25,
        'J/psi': 443,
        'KL': 130,
        'KS': 310,
        'K*+': 323,
        'K*0': 313,
        'K+': 321,
        'K0': 311,
        'Lambda': 3122,
        'Lambdab': 5122,
        'Lambdac': 4122,
        'omega': 223,
        'D*0': 423,
        'D*+': 413,
        'W': 24,
        'Z': 23,
        'e': 11,
        'eta': 221,
        'f0': 9010221,
        'mu': 13,
        'phi': 333,
        'pi+': 211,
        'pi0': 111,
        'psi(2S)': 100443,
        'rho+': 213,
        'rho0': 113,
        't': 6,
        'tau': 15,
        'u': 2,
        'p': 2212,
        'n': 2112,
    }
    _pdg_particles_inv = {v:k for k,v in PDG_PARTICLES.items()}
    _pdg_tex_regex = re.compile(
        r"^([A-Za-z\\/]+)" # latin or greek letters or slash
        r"(?:_\{(.*?)\})*" # _{...}
        r"(?:\^\{(.*?)\})*" # ^{...}
        r"(?:\((.*?)\))*" # (...)
        r"(?:\^\{(.*?)\})*" # ^{...}
    )

    @classmethod
    def from_flavio_name(cls, flavio_name):
        return cls.from_pdgid(cls.PDG_PARTICLES[flavio_name])

    @classmethod
    def flavio_all(cls):
        return {particle for particle in cls.all() if particle.flavio_name}

    @property
    def flavio_name(self):
        return self._pdg_particles_inv.get(self.pdgid, None)

    @property
    def latex_name_simplified(self):
        m = self._pdg_tex_regex.match(self.latex_name)
        if m is None:
            return self.latex_name
        name = m.group(1)
        sub = m.group(2)
        sup = (m.group(3) or '') + (m.group(5) or '')
        par = m.group(4)
        if sub or name in ('W', 'Z', 'H', 'e', '\\mu', '\\tau'):
            # remove superscripts +-0 and keep only *
            sup = '*' if '*' in sup else ''
        if not sub and par and not par.isdigit() and name != 'J/\\psi':
            # subscript absent and parantheses contain letter but not for 'J/\\psi'
            sub = par
        sub_tex = r'_{' + sub + r'}' if sub else ''
        sup_tex = r'^{' + sup + r'}' if sup else ''
        return name + sub_tex + sup_tex

    @property
    def flavio_m(self):
        name = 'm_' + self.flavio_name
        tex = r'$m_{' + self.latex_name_simplified + '}$'
        pole_mass = ' quark pole' if self.name == 't' else ''
        description = r'${}${} mass'.format(
            self.latex_name_simplified, pole_mass
        )
        central = self.mass*1e-3
        right = self.mass_upper*1e-3
        left = self.mass_lower*1e-3
        return name, tex, description, central, right, left

    @property
    def flavio_tau(self):
        if {self.width, self.width_upper, self.width_lower} & {None, 0}:
            return None
        name = 'tau_' + self.flavio_name
        tex = r'$\tau_{' + self.latex_name_simplified + '}$'
        description = r'${}$ lifetime'.format(self.latex_name_simplified)
        G_central = self.width*1e-3
        G_right = self.width_upper*1e-3
        G_left = self.width_lower*1e-3
        central = 1/G_central # life time = 1/width
        right = G_right/G_central**2
        left = G_left/G_central**2
        return name, tex, description, central, right, left

def read_pdg(year, constraints):
    """Read particle masses and widths from the PDG data file of a given year."""
    FlavioParticle.load_table(p_data.open_text(p_data, "particle{}.csv".format(year)))
    for particle in FlavioParticle.flavio_all():
        for data in (particle.flavio_m, particle.flavio_tau):
            if data is None:
                continue
            name, tex, description, central, right, left = data
            try:
                # if parameter already exists, remove existing constraints on it
                p = Parameter[name]
                constraints.remove_constraint(name)
            except KeyError:
                # otherwise, create it
                p = Parameter(name)
            p.tex = tex
            p.description = description
            if right == left:
                constraints.add_constraint([name],
                    NormalDistribution(central, right))
            else:
                constraints.add_constraint([name],
                    AsymmetricNormalDistribution(central,
                    right_deviation=right, left_deviation=left))



############### Read default parameters ###################

# Create the object
default_parameters = ParameterConstraints()

# read default parameters
default_parameters.read_default()



r"""Functions for lepton flavour violating $\tau\to \ell\gamma$ decays."""

import flavio
from flavio.physics.taudecays.taulgamma import BR_llgamma


def BR_muegamma(wc_obj, par):
    r"""Branching ratio of $\mu\to e\gamma$."""
    return BR_llgamma(wc_obj, par, 'mu', 'e')


_process_tex = r"\mu\to e\gamma"
_process_taxonomy = r'Process :: muon decays :: LFV decays :: $' + _process_tex + r"$"

_obs_name = "BR(mu->egamma)"
_obs = flavio.classes.Observable(_obs_name)
_obs.set_description(r"Branching ratio of $" + _process_tex + r"$")
_obs.tex = r"$\text{BR}(" + _process_tex + r")$"
_obs.add_taxonomy(_process_taxonomy)
flavio.classes.Prediction(_obs_name, BR_muegamma)



import unittest
import flavio
import wilson
from math import sqrt


par = flavio.default_parameters.get_central_all()
wc = flavio.WilsonCoefficients()
wc.set_initial({}, 91.1876, 'SMEFT', 'Warsaw')


class TestHiggsProduction(unittest.TestCase):
    def test_mugg(self):
        RSM = flavio.physics.higgs.production.ggF(wc.wc)
        self.assertEqual(RSM, 1)


class TestHiggsDecay(unittest.TestCase):
    def test_mugg(self):
        RSM = flavio.physics.higgs.decay.h_bb(wc.wc)
        self.assertEqual(RSM, 1)


class TestHiggsWidth(unittest.TestCase):
    def test_gamma_h(self):
        RSM = flavio.physics.higgs.width.Gamma_h(par, wc.wc)
        self.assertEqual(RSM, 1)

    def test_gamma_h_quadratic(self):
        # large up quark Yukawa coupling to make h>uu comparable to h>bb
        C = -sqrt(2) * 2.80 / 246.22**3  # -sqrt(2) * m_b(m_h) / v**2
        wc_np = wilson.Wilson({'uphi_11': C}, 125, 'SMEFT',  'Warsaw')
        R = flavio.physics.higgs.width.Gamma_h(par, wc_np.wc)
        self.assertAlmostEqual(R, 1.58, delta=0.15)


class TestHiggsSignalStrengths(unittest.TestCase):
    def test_mugg(self):
        muSM = flavio.physics.higgs.signalstrength.higgs_signalstrength(wc, par, 'ggF', 'h_tautau')
        self.assertEqual(muSM, 1)

    def test_obs_sm_ff(self):
        muSM = flavio.sm_prediction('mu_gg(h->tautau)')
        self.assertEqual(muSM, 1)

    def test_obs_sm_gaga(self):
        muSM = flavio.sm_prediction('mu_gg(h->gammagamma)')
        self.assertEqual(muSM, 1)

    def test_obs_np_ff(self):
        w = wilson.Wilson({'ephi_33': 1e-8}, 125, 'SMEFT', 'Warsaw')
        mu_tau = flavio.np_prediction('mu_gg(h->tautau)', w)
        mu_mu = flavio.np_prediction('mu_gg(h->mumu)', w)
        # mu_mu is not exactly =1 since it modifies the Higgs total width
        self.assertAlmostEqual(mu_mu, 1, delta=0.02)
        self.assertNotAlmostEqual(mu_tau, 1, delta=0.02)

    def test_obs_np_gaga(self):
        w = wilson.Wilson({'phiW': 1e-6}, 125, 'SMEFT', 'Warsaw')
        mu = flavio.np_prediction('mu_gg(h->gammagamma)', w)
        self.assertNotAlmostEqual(mu, 1, delta=0.02)

    def test_noerror(self):
        flavio.sm_prediction('mu_gg(h->Zgamma)')
        flavio.sm_prediction('mu_tth(h->tautau)')
        flavio.sm_prediction('mu_Wh(h->mumu)')
        flavio.sm_prediction('mu_Zh(h->bb)')
        flavio.sm_prediction('mu_VBF(h->WW)')


class TestHiggsMeasurements(unittest.TestCase):
    def test_run1_correlation(self):
        # check that the large anticorrlation between Wh and Zh gaga is correct
        m = flavio.classes.Measurement['LHC Run 1 Higgs combination']
        c = m._constraints[0][0]
        self.assertEqual(len(c.central_value), 20)
        self.assertAlmostEqual(c.correlation[12, 8], -0.64, delta=0.005)
        self.assertEqual(m.all_parameters[8], 'mu_Wh(h->gammagamma)')
        self.assertEqual(m.all_parameters[12], 'mu_Zh(h->gammagamma)')

    def test_atlas_run2_correlation(self):
        # check that the large anticorrlation between ggF and VBF tautau is correct
        m = flavio.classes.Measurement['ATLAS Run 2 Higgs 80/fb']
        c = m._constraints[0][0]
        self.assertEqual(len(c.central_value), 16)
        self.assertAlmostEqual(c.correlation[3, 7], -0.44, delta=0.005)
        self.assertEqual(m.all_parameters[3], 'mu_gg(h->tautau)')
        self.assertEqual(m.all_parameters[7], 'mu_VBF(h->tautau)')


    def test_cms_run2_correlation(self):
        # check that the large anticorrlation between ggF and VBF mumu is correct
        m = flavio.classes.Measurement['CMS Run 2 Higgs 36/fb']
        c = m._constraints[0][0]
        self.assertEqual(len(c.central_value), 24)
        self.assertAlmostEqual(c.correlation[5, 10], -0.54, delta=0.005)
        self.assertEqual(m.all_parameters[5], 'mu_gg(h->mumu)')
        self.assertEqual(m.all_parameters[10], 'mu_VBF(h->mumu)')



"""Functions for conversion of quark masses not contained in RunDec."""

from math import log, pi
import numpy as np
import flavio
from flavio.math.functions import zeta
from flavio.physics.running.masses import zeta


# (19) of arXiv:1107.3100v1
def fKsFromMs1(Mu, M, Nf):
    flavio.citations.register("Gambino:2011cq")
    return -(4/3.)* (1- (4/3) * (Mu/M) - (Mu**2/(2*M**2)) )
def fKsFromMs2(Mu, M, Nf):
    flavio.citations.register("Gambino:2011cq")
    b0 = 11 - 2*Nf/3.
    return ((((1/3.)*log(M/(2*Mu))+13/18.)*b0 - pi**2/3. + 23/18.)*Mu**2/M**2
            + (((8/9.)*log(M/(2*Mu))+64/27.)*b0 - 8*pi**2/9. + 92/27.) * Mu/M
            - (pi**2/12. + 71/96.)*b0
            + zeta(3)/6. - pi**2/9. * log(2) + 7*pi**2/12. + 23/72.
            )
# from (A.8) of hep-ph/0302262v1
def fKsFromMs3(Mu, M, Nf):
    flavio.citations.register("Benson:2003kp")
    b0 = 11 - 2*Nf/3.
    return -(b0/2.)**2*(2353/2592.+13/36.*pi**2+7/6.*zeta(3)
            -16/9.*Mu/M*((log(M/(2*Mu))+8/3.)**2+67/36.-pi**2/6.)
            -2/3.*Mu**2/M**2*((log(M/(2*Mu))+13/6.)**2+10/9.-pi**2/6.))

def mKS2mMS(M, Nf, asM, Mu, nl):
    s = np.zeros(4)
    s[0] = 1.
    s[1] = (asM/pi) * fKsFromMs1(Mu, M, Nf)
    s[2] = (asM/pi)**2 * fKsFromMs2(Mu, M, Nf)
    s[3] = (asM/pi)**3 * fKsFromMs3(Mu, M, Nf)
    r = s[:nl+1].sum()
    return M * r

def mMS2mKS(MS, Nf, asM, Mu, nl):
    def convert(M):
        s = np.zeros(4)
        s[0] = 1.
        s[1] = -(asM/pi) * fKsFromMs1(Mu, M, Nf)
        s[2] = -(asM/pi)**2 * fKsFromMs2(Mu, M, Nf)
        # properly invert the relation to O(asM**2)
        s[2] = s[2] + s[1]**2
        s[3] = -(asM/pi)**3 * fKsFromMs3(Mu, M, Nf)
        r = s[:nl+1].sum()
        return MS * r
    # iterate twice
    Mtmp = convert(MS)
    Mtmp = convert(Mtmp)
    return convert (Mtmp)



import unittest
import flavio
from flavio.classes import AuxiliaryQuantity

constraints = flavio.default_parameters
wc_obj = flavio.WilsonCoefficients()
par = constraints.get_central_all()


class TestDPFF(unittest.TestCase):
    def test_etm(self):
        # compare to ETM result at kinematic endpoints
        q2max_pi = (par['m_D0'] - par['m_pi+'])**2
        q2max_K = (par['m_D0'] - par['m_K+'])**2
        ff_pi_0 = AuxiliaryQuantity["D->pi form factor"].prediction_central(constraints, wc_obj, 0)
        ff_K_0 = AuxiliaryQuantity["D->K form factor"].prediction_central(constraints, wc_obj, 0)
        ff_pi_max = AuxiliaryQuantity["D->pi form factor"].prediction_central(constraints, wc_obj, q2max_pi)
        ff_K_max = AuxiliaryQuantity["D->K form factor"].prediction_central(constraints, wc_obj, q2max_K)
        self.assertAlmostEqual(ff_pi_0['f0'], 0.612, delta=0.001)
        self.assertAlmostEqual(ff_pi_max['f0'], 1.134, delta=0.01)
        self.assertAlmostEqual(ff_pi_0['f+'], 0.612, delta=0.001)
        self.assertAlmostEqual(ff_pi_max['f+'], 2.130, delta=0.02)
        self.assertAlmostEqual(ff_pi_0['fT'], 0.506, delta=0.03)
        self.assertAlmostEqual(ff_pi_max['fT'], 1.573, delta=0.03)
        self.assertAlmostEqual(ff_K_0['f0'], 0.765, delta=0.001)
        self.assertAlmostEqual(ff_K_max['f0'], 0.979, delta=0.001)
        self.assertAlmostEqual(ff_K_0['f+'], 0.765, delta=0.001)
        self.assertAlmostEqual(ff_K_max['f+'], 1.336, delta=0.01)
        self.assertAlmostEqual(ff_K_0['fT'], 0.687, delta=0.001)
        self.assertAlmostEqual(ff_K_max['fT'], 1.170, delta=0.01)



import unittest
import numpy as np
from . import wilsoncoefficients
from .. import eft
import flavio

s = 1.519267515435317e+24

par = {
    'm_e': 0.510998928e-3,
    'm_mu': 105.6583715e-3,
    'm_tau': 1.77686,
    'm_B+': 5.27929,
    'm_B0': 5.27961,
    'm_Bs': 5.36679,
    'm_K*0': 0.89166,
    'tau_B+': 1638.e-15*s,
    'tau_B0': 152.e-14*s,
    'alpha_e': 1/127.940,
    'alpha_s': 0.1185,
    'm_Z': 91.1876,
    'm_b': 4.17,
    'm_s': 0.1,
    'm_t': 173.1,
    'm_c': 1.275,
    'GF': 1.1663787e-5,
    'Vus': 0.22,
    'Vub': 3.7e-3,
    'Vcb': 4.1e-2,
    'delta': 1.22,
}

class TestBWilson(unittest.TestCase):
    def test_wctot(self):
        wc_low_correct = np.array([ -2.93671059e-01,   1.01676402e+00,  -5.87762813e-03,
        -8.70666812e-02,   4.11098919e-04,   1.10641294e-03,
        -2.95662859e-01,  -1.63048361e-01,   4.11363023e+00,
        -4.19345312e+00,   3.43507549e-03,   1.22202095e-03,
        -1.03192325e-03,  -1.00703396e-04,  -3.17810374e-03])
        wc_obj = eft.WilsonCoefficients()
        wc_low = wilsoncoefficients.wctot_dict(wc_obj, 'bsmumu', 4.2, par)
        wc_names = ['C1_bs', 'C2_bs', 'C3_bs', 'C4_bs', 'C5_bs', 'C6_bs', 'C7_bs', 'C8_bs', 'C9_bsmumu', 'C10_bsmumu', 'C3Q_bs', 'C4Q_bs', 'C5Q_bs', 'C6Q_bs', 'Cb_bs', 'C1p_bs', 'C2p_bs', 'C3p_bs', 'C4p_bs', 'C5p_bs', 'C6p_bs', 'C7p_bs', 'C8p_bs', 'C9p_bsmumu', 'C10p_bsmumu', 'C3Qp_bs', 'C4Qp_bs', 'C5Qp_bs', 'C6Qp_bs', 'Cbp_bs', 'CS_bsmumu', 'CP_bsmumu', 'CSp_bsmumu', 'CPp_bsmumu']
        wc_low_array = np.asarray([wc_low[key] for key in wc_names])
        yi = np.array([0, 0, -1/3., -4/9., -20/3., -80/9.])
        zi = np.array([0, 0, 1, -1/6., 20, -10/3.])
        wc_low_array[6] = wc_low_array[6] + np.dot(yi, wc_low_array[:6]) # c7eff
        wc_low_array[7] = wc_low_array[7] + np.dot(zi, wc_low_array[:6]) # c8eff
        np.testing.assert_almost_equal(wc_low_array[:15], wc_low_correct, decimal=2)

    def test_C78p(self):
        wc_obj = eft.WilsonCoefficients()
        wc_low = wilsoncoefficients.wctot_dict(wc_obj, 'bsmumu', 4.2, par)
        ms = flavio.physics.running.running.get_ms(par, 4.2, nf_out=5)
        mb = flavio.physics.running.running.get_mb(par, 4.2, nf_out=5)
        self.assertAlmostEqual(wc_low['C7p_bs']/wc_low['C7_bs'], ms/mb)
        self.assertAlmostEqual(wc_low['C8p_bs']/wc_low['C8_bs'], ms/mb)

    def test_clnu(self):
        par_dict = flavio.default_parameters.get_central_all()
        par_dict['alpha_s'] = 0.1184
        par_dict['alpha_e'] = 1/127.925
        par_dict['s2w']  = 0.2315
        par_dict['m_t']  = 173.3
        cl = wilsoncoefficients.CL_SM(par_dict)
        # comparing the central value of X_t to (4.2) of 1009.0947
        self.assertAlmostEqual(-cl*par_dict['s2w']/1.469, 1, delta=0.01)



import flavio
from collections import Counter
import warnings
import numpy as np
import voluptuous as vol
import pickle
from flavio.classes import NamedInstanceClass
from flavio.statistics.probability import NormalDistribution, MultivariateNormalDistribution
from flavio.io import instanceio as iio


class MeasurementLikelihood(iio.YAMLLoadable):
    """A `MeasurementLikelihood` provides a likelihood function from
    experimental measurements.

    Methods:

    - `get_predictions_par`: Return a dictionary of SM predictions for the
    observables of interest
    - `log_likelihood_pred`: The likelihood as a function of the predictions
    - `log_likelihood_par`: The likelihood as a function of parameters and
    Wilson coefficients

    Instances can be imported and exported from/to YAML using the `load`
    and `dump` methods.
    """

    _input_schema_dict = {
        'observables':  vol.All([iio.coerce_observable_tuple], iio.list_deduplicate),
        'exclude_measurements': vol.Any(None, [str]),
        'include_measurements': vol.Any(None, [str]),
    }

    _output_schema_dict = {
        'observables':  [iio.coerce_observable_dict],
        'exclude_measurements': vol.All(iio.ensurelist, [str]),
        'include_measurements': vol.All(iio.ensurelist, [str]),
    }

    def __init__(self, observables, *,
                 exclude_measurements=None,
                 include_measurements=None,
                 include_pseudo_measurements=False):
        """Initialize the instance.

        Parameters:
        - `observables`: list of observables (tuples or strings)
        - `exclude_measurements`: list of measurement names to exclude
        (default: none ar excluded)
        - `include_measurements`: list of measurement names to include
        (default: all are included)

        Only one of `exclude_measurements` or `include_measurements` should
        be specified. By default, all existing instances of
        `flavio.Measurements` are included as constraints (except if they
        carry 'Pseudo-measurement' in their name).
        """
        super().__init__()
        self.observables = observables
        self.exclude_measurements = exclude_measurements
        self.include_measurements = include_measurements
        self.include_pseudo_measurements = include_pseudo_measurements
        self._predictions_cache_hash = None
        self._predictions_cache_values = None
        if exclude_measurements and include_measurements:
            raise ValueError("The options exclude_measurements and include_measurements must not be specified simultaneously")

        # check that observables are constrained
        _obs_measured = set()
        for m_name in self.get_measurements:
            m_obj = flavio.Measurement[m_name]
            _obs_measured.update(m_obj.all_parameters)
        missing_obs = set(observables) - set(_obs_measured).intersection(set(observables))
        assert missing_obs == set(), "No measurement found for the observables: " + str(missing_obs)
        self._warn_meas_corr()

    def _warn_meas_corr(self):
        """Warn the user if the fit contains multiple correlated measurements of
        an observable that is not included in the fit parameters, as this will
        lead to inconsistent results."""
        corr_with = {}
        # iterate over all measurements constraining at least one fit obs.
        for name in self.get_measurements:
            m = flavio.classes.Measurement[name]
            # iterate over all fit obs. constrained by this measurement
            for obs in set(self.observables) & set(m.all_parameters):
                # the constraint on this fit obs.
                constraint = m._parameters[obs][1]
                # find all the other obs. constrained by this constraint
                for c, p in m._constraints:
                    if c == constraint:
                        par = p
                        break
                for p in par:
                    # if the other obs. are not fit obs., append them to the list
                    if p not in self.observables:
                        if p not in corr_with:
                            corr_with[p] = [obs]
                        else:
                            corr_with[p].append(obs)
        # replace list by a Counter
        corr_with = {k: Counter(v) for k, v in corr_with.items() if v}
        # warn for all counts > 1
        for obs1, counter in corr_with.items():
            for obs2, count in counter.items():
                if count > 1:
                    warnings.warn(("{} of the measurements in the likelihood "
                                   "constrain both '{}' and '{}', but only the "
                                   "latter is included among the fit "
                                   "observables. This can lead to inconsistent "
                                   "results as the former is profiled over."
                                   ).format(count, obs1, obs2))
        return corr_with

    @property
    def get_measurements(self):
        """Return a list of all the measurements currently defined that
        constrain any of the fit observables."""
        all_measurements = []
        for m_name, m_obj in flavio.classes.Measurement.instances.items():
            if m_name.split(' ')[0] == 'Pseudo-measurement' and not self.include_pseudo_measurements:
                # skip pseudo measurements generated by FastFit instances
                continue
            if set(m_obj.all_parameters).isdisjoint(self.observables):
                # if set of all observables constrained by measurement is disjoint
                # with fit observables, do nothing
                continue
            else:
                # else, add measurement name to output list
                all_measurements.append(m_name)
        if self.exclude_measurements is None and self.include_measurements is None:
            return all_measurements
        elif self.exclude_measurements is not None:
            return list(set(all_measurements) - set(self.exclude_measurements))
        elif self.include_measurements is not None:
            return list(set(all_measurements) & set(self.include_measurements))

    def get_predictions_par(self, par_dict, wc_obj):
        """Compute the predictions for all observables as functions of
        a parameter dictionary `par_dict`and WilsonCoefficient instance
        `wc_obj`.
        The latest computed values are cached and returned if the function is
        called successively with the same arguments.
        """
        # compute a hash from the function's arguments used for caching
        arg_hash = hash((frozenset(par_dict.items()),wc_obj))
        if arg_hash != self._predictions_cache_hash:
            all_predictions = {}
            for observable in self.observables:
                obs = flavio.classes.Observable.argument_format(observable, 'dict')
                name = obs.pop('name')
                _inst = flavio.classes.Observable[name]
                all_predictions[observable] = _inst.prediction_par(par_dict, wc_obj, **obs)
            self._predictions_cache_values = all_predictions
            self._predictions_cache_hash = arg_hash
        return self._predictions_cache_values

    def get_number_observations(self, exclude_observables=None):
        """Get the number of observations, defined as individual measurements
        of observables."""
        nobs = 0
        for measurement in self.get_measurements:
            m_obj = flavio.Measurement[measurement]
            m_obs = set(m_obj.all_parameters)
            exclude_observables = set(exclude_observables or [])
            our_obs =  set(self.observables) - exclude_observables
            # add the number of observables contained in the measurement
            # and in our instance (except the excluded ones)
            nobs += len(m_obs & our_obs)
        return nobs


    def log_likelihood_pred(self, pred_dict, exclude_observables=None, delta=False):
        """Return the logarithm of the likelihood function as a function of
        a dictionary of observable predictions `pred_dict`"""
        ll = 0.
        for measurement in self.get_measurements:
            m_obj = flavio.Measurement[measurement]
            m_obs = set(m_obj.all_parameters)
            exclude_observables = set(exclude_observables or [])
            # exclude the observables not contained in the likelihood
            # as well as the ones explicitly excluded by `exclude_observables`
            exclude_parameters = m_obs - set(self.observables) | exclude_observables
            if m_obs == exclude_parameters:
                continue  # if all measured observables are excluded: nothing to do
            prob_dict = m_obj.get_logprobability_all(pred_dict, exclude_parameters=exclude_parameters, delta=delta)
            ll += sum(prob_dict.values())
        return ll

    def log_likelihood_par(self, par_dict, wc_obj, exclude_observables=None, delta=False):
        """Return the logarithm of the likelihood function as a function of
        a parameter dictionary `par_dict` and WilsonCoefficient instance
        `wc_obj`"""
        predictions = self.get_predictions_par(par_dict, wc_obj)
        return self.log_likelihood_pred(predictions, exclude_observables=exclude_observables, delta=delta)


class ParameterLikelihood(iio.YAMLLoadable):
    """A `ParameterLikelihood` provides a likelihood function in terms of
    parameters.

    Methods:

    - `log_likelihood_par`: The likelihood as a function of the parameters
    - `get_central`: get an array with the parameters' central values
    - `get_random`: get an array with random values for the parameters

    Instances can be imported and exported from/to YAML using the `load`
    and `dump` methods.
    """
    _input_schema_dict = {
        'par_obj': vol.All([dict], iio.coerce_par_obj),
        'parameters': vol.Any(None, [str]),
    }

    _output_schema_dict = {
        'par_obj': iio.get_par_diff,
        'parameters': vol.Any(iio.ensurelist, [str]),
    }

    def __init__(self,
                 par_obj=flavio.default_parameters,
                 parameters=None):
        """Initialize the instance.

        Parameters:

        - `par_obj`: an instance of `ParameterConstraints` (defaults to
        `flavio.default_parameters`)
        - parameters: a list of parameters whose constraints should be taken
        into account in the likelihood.
        """
        self.par_obj = par_obj
        self.parameters = parameters
        self.parameters_central = self.par_obj.get_central_all()

    def log_likelihood_par(self, par_dict, delta=False):
        """Return the prior probability for all parameters.

        Note that only the parameters in `self.parameters` will give a
        contribution to the likelihood."""
        exclude_parameters = list(set(self.par_obj._parameters.keys())-set(self.parameters))
        prob_dict = self.par_obj.get_logprobability_all(par_dict, exclude_parameters=exclude_parameters, delta=delta)
        return sum([p for obj, p in prob_dict.items()])

    @property
    def get_central(self):
        """Return a numpy array with the central values of all parameters."""
        return np.asarray([self.parameters_central[p] for p in self.parameters])

    @property
    def get_random(self):
        """Return a numpy array with random values for all parameters."""
        all_random = self.par_obj.get_random_all()
        return np.asarray([all_random[p] for p in self.parameters])


class Likelihood(iio.YAMLLoadable):
    """A `Likelihood` provides a likelihood function consisting of a
    contribution from experimental measurements and a contribution from
    parameters.

    Methods:

    - `log_prior_fit_parameters`: The parameter contribution to the
    log-likelihood
    - `log_likelihood_exp`: The experimental contribution to the
    log-likelihood
    - `log_likelihood`: The total log-likelihood that is the sum of the
    parameter and the experimental contribution

    Instances can be imported and exported from/to YAML using the `load`
    and `dump` methods.
    """

    _input_schema_dict = {
        'par_obj': vol.All([dict], iio.coerce_par_obj),
        'fit_parameters': vol.Any(iio.ensurelist, [str]),
        'observables':  vol.All([iio.coerce_observable_tuple], iio.list_deduplicate),
        'exclude_measurements': vol.Any(iio.ensurelist, [str]),
        'include_measurements': vol.Any(iio.ensurelist, [str]),
    }

    _output_schema_dict = {
        'par_obj': vol.All([dict], iio.coerce_par_obj),
        'fit_parameters': iio.get_par_diff,
        'observables':  [iio.coerce_observable_dict],
        'exclude_measurements': vol.Any(iio.ensurelist, [str]),
        'include_measurements': vol.Any(iio.ensurelist, [str]),
    }

    def __init__(self,
                 par_obj=flavio.default_parameters,
                 fit_parameters=None,
                 observables=None,
                 exclude_measurements=None,
                 include_measurements=None,
                 include_pseudo_measurements=False,
                 ):
        self.par_obj = par_obj
        self.parameters_central = self.par_obj.get_central_all()
        self.fit_parameters = fit_parameters or []
        self.observables = observables
        self.exclude_measurements = exclude_measurements
        self.include_measurements = include_measurements
        self.measurement_likelihood = MeasurementLikelihood(
            observables,
            exclude_measurements=exclude_measurements,
            include_measurements=include_measurements,
            include_pseudo_measurements=include_pseudo_measurements,)
        self.parameter_likelihood = ParameterLikelihood(
            par_obj=par_obj,
            parameters=fit_parameters)

    def log_prior_fit_parameters(self, par_dict, delta=False):
        """Parameter contribution to the log-likelihood."""
        if not self.fit_parameters:
            return 0  # nothing to do
        return self.parameter_likelihood.log_likelihood_par(par_dict, delta=delta)

    def log_likelihood_exp(self, par_dict, wc_obj, exclude_observables=None, delta=False):
        """Experimental contribution to the log-likelihood."""
        return self.measurement_likelihood.log_likelihood_par(par_dict, wc_obj, exclude_observables=exclude_observables, delta=delta)

    def log_likelihood(self, par_dict, wc_obj, exclude_observables=None, delta=False):
        """Total log-likelihood.

        Parameters:
        - `par_dict`: a dictionary of parameter values
        - `wc_obj`: an instance of `WilsonCoefficients` or `wilson.Wilson`
        """
        return self.log_prior_fit_parameters(par_dict, delta=delta) + self.log_likelihood_exp(par_dict, wc_obj, exclude_observables=exclude_observables, delta=delta)


class SMCovariance(object):
    """Class to compute, save, and load a covariance matrix of SM
    predictions.

    Methods:

    - `compute`: Compute the covariance
    - `get`: Compute the covariance if necessary, otherwise return cached one
    - `save`: Save the covariance to a file
    - `load`: Load the covariance from a file
    - `load_dict`: Load the covariance from a dictionary
    """

    def __init__(self, observables, *,
                 vary_parameters='all', par_obj=None):
        """Initialize the class.

        Parameters:
        - `observables`: list of observables
        - `vary_parameters`: parameters to vary. Defaults to 'all'.
        - `par_obj`: instance of ParameterConstraints. Defaults to
        flavio.default_parameters.
        """
        self.observables = observables
        self.vary_parameters = vary_parameters
        self.par_obj = par_obj or flavio.default_parameters
        self._cov = None

    def compute(self, N, threads):
        """Compute the covariance for `N` random values, using `threads`
        CPU threads."""
        return flavio.sm_covariance(obs_list=self.observables,
                                    N=N,
                                    par_vary=self.vary_parameters,
                                    par_obj=self.par_obj,
                                    threads=threads)

    def get(self, N=100, threads=1, force=True):
        """Compute the covariance for `N` random values (default: 100),
        using `threads` CPU threads (default: 1).

        If `force` is False, return a cached version if it exists.
        """
        if self._cov is None or force:
            self._cov = self.compute(N=N, threads=threads)
        elif N != 100:
            warnings.warn("Argument N={} ignored ".format(N) + \
                          "as covariance has already been " + \
                          "computed. Recompute using `force=True`.")
        return self._cov

    def save(self, filename):
        """Save the SM covariance to a pickle file.

        The covariance must have been computed before using `get` or
        `compute`.
        """
        if self._cov is None:
            raise ValueError("Call `get` or `compute` first.")
        with open(filename, 'wb') as f:
            data = dict(covariance=self._cov,
                        observables=self.observables)
            pickle.dump(data, f)

    def load(self, filename):
        """Load the SM covariance from a pickle file."""
        with open(filename, 'rb') as f:
            data = pickle.load(f)
        self.load_dict(d=data)

    def load_dict(self, d):
        """Load the SM covariance from a dictionary.

        It must have the form `{'observables': [...], 'covariance': [[...]]}`
        where 'covariance' is a covariance matrix in the basis of observables
        given by 'observables' which must at least contain all the observables
        involved in the fit. Additional observables will be ignored; the
        ordering is arbitrary.
        """
        obs = d['observables']
        try:
            permutation = [obs.index(o) for o in self.observables]
        except ValueError:
            raise ValueError("Covariance matrix does not contain all necessary entries")
        assert len(permutation) == len(self.observables), \
            "Covariance matrix does not contain all necessary entries"
        if len(permutation) == 1:
            if d['covariance'].shape == ():
                self._cov = d['covariance']
            else:
                self._cov = d['covariance'][permutation][:,permutation][0,0]
        else:
            self._cov = d['covariance'][permutation][:,permutation]


class MeasurementCovariance(object):
    """Class to compute, save, and load a covariance matrix and the central
    values of experimental measurements.

    Methods:

    - `compute`: Compute the covariance
    - `get`: Compute the covariance if necessary, otherwise return cached one
    - `save`: Save the covariance to a file
    - `load`: Load the covariance from a file
    - `load_dict`: Load the covariance from a dictionary
    """

    def __init__(self, measurement_likelihood):
        """Initialize the class.

        Parameters:
        - `measurement_likelihood`: an instance of `MeasurementLikelihood`
        """
        self.measurement_likelihood = measurement_likelihood
        self._central_cov = None

    def compute(self, N):
        """Compute the covariance for `N` random values."""
        ml = self.measurement_likelihood
        means = []
        covariances = []
        for measurement in ml.get_measurements:
            m_obj = flavio.Measurement[measurement]
            # obs. included in the fit and constrained by this measurement
            our_obs = set(m_obj.all_parameters).intersection(ml.observables)
            # construct a dict. containing a vector of N random values for
            # each of these observables
            random_dict = m_obj.get_random_all(size=N)
            random_arr = np.zeros((len(ml.observables), N))
            for i, obs in enumerate(ml.observables):
                if obs in our_obs:
                    random_arr[i] = random_dict[obs]
            mean = np.mean(random_arr, axis=1)
            covariance = np.cov(random_arr)
            for i, obs in enumerate(ml.observables):
                if obs not in our_obs:
                    covariance[:,i] = 0
                    covariance[i, :] = 0
                    covariance[i, i] = np.inf
            means.append(mean)
            covariances.append(covariance)
        # if there is only a single measuement
        if len(means) == 1:
            if len(means[0]) == 1:
                # if there is only a single observable
                return means[0][0], covariances[0]
            else:
                return means[0], covariances[0]
        # if there are severeal measurements, perform a weighted average
        else:
            # covariances: [Sigma_1, Sigma_2, ...]
            # means: [x_1, x_2, ...]
            # weights_ [W_1, W_2, ...] where W_i = (Sigma_i)^(-1)
            # weighted covariance is  (W_1 + W_2 + ...)^(-1) = Sigma
            # weigted mean is  Sigma.(W_1.x_1 + W_2.x_2 + ...) = x
            if len(ml.observables) == 1:
                weights = np.array([1/c for c in covariances])
                weighted_covariance = 1/np.sum(weights, axis=0)
                weighted_mean = weighted_covariance * np.sum(
                                [np.dot(weights[i], means[i]) for i in range(len(means))])
            else:
                weights = [np.linalg.inv(c) for c in covariances]
                weighted_covariance = np.linalg.inv(np.sum(weights, axis=0))
                weighted_mean = np.dot(weighted_covariance, np.sum(
                                [np.dot(weights[i], means[i]) for i in range(len(means))],
                                axis=0))
            return weighted_mean, weighted_covariance

    def get(self, N=5000, force=True):
        """Compute the covariance for `N` random values (default: 5000).

        If `force` is False, return a cached version if it exists.
        """
        if self._central_cov is None or force:
            self._central_cov = self.compute(N=N)
        elif N != 5000:
            warnings.warn("Argument N={} ignored ".format(N) + \
                          "as experimental covariance has already been " + \
                          "computed. Recompute using `force=True`.")
        return self._central_cov

    def save(self, filename):
        """Save the experimental central values and the covariance to a pickle
        file.

        The central values and the covariance must have been computed before
        using `get` or `compute`."""
        if self._central_cov is None:
            raise ValueError("Call `get` or `compute` first.")
        with open(filename, 'wb') as f:
            data = dict(central=self._central_cov[0],
                        covariance=self._central_cov[1],
                        observables=self.measurement_likelihood.observables)
            pickle.dump(data, f)

    def load(self, filename):
        """Load the experimental central values and the covriance from a pickle
        file."""
        with open(filename, 'rb') as f:
            data = pickle.load(f)
        self.load_dict(d=data)

    def load_dict(self, d):
        """Load the the experimental central values and the covariance from a
        dictionary.

        It must have the form
        `{'observables': [...], 'central': [...], 'covariance': [[...]]}`
        where 'central' is a vector of central values and 'covariance' is a
        covariance matrix, both in the basis of observables given by
        'observables' which must at least contain all the observables
        involved in the fit. Additional observables will be ignored; the
        ordering is arbitrary."""
        ml = self.measurement_likelihood
        obs = d['observables']
        try:
            permutation = [obs.index(o) for o in ml.observables]
        except ValueError:
            raise ValueError("Covariance matrix does not contain all necessary entries")
        assert len(permutation) == len(ml.observables), \
            "Covariance matrix does not contain all necessary entries"
        if np.isscalar(d['central']):
            self._central_cov = (
                d['central'],
                d['covariance']
            )
        else:
            self._central_cov = (
                d['central'][permutation],
                d['covariance'][permutation][:, permutation],
            )


class FastLikelihood(NamedInstanceClass, iio.YAMLLoadable):
    """A variant (but not subclass) of `Likelihood` where some or all
    of the parameters have been "integrated out" and their theoretical
    uncertainties combined with the experimental uncertainties into
    a multivariate Gaussian "pseudo measurement".

    The pseuo measurement is generated by calling the method `make_measurement`.
    This is done by
    generating random samples of the nuisance parameters and evaluating all
    observables within the Standard Model many times (100 by default).
    Then, the covariance of all predictions is extracted. Similarly, a covariance
    matrix for all experimental measurements is determined. Both covariance
    matrices are added and the resulting multivariate Gaussian treated as a
    single measurement.

    This approach has the advantage that two-dimensional plots of the likelihood
    can be produced without the need for sampling or profiling the other
    dimensions. However, several strong assumptions go into this method, most
    importantly,

    - all uncertainties - experimental and theoretical - are treated as Gaussian
    - the theoretical uncertainties in the presence of new physics are assumed
      to be equal to the ones in the SM

    Methods:

    - `make_measurement`: Generate the pseudo measurement
    - `log_likelihood`: The log-likelihood function

    Important attributes/properties:

    - `likelihood`: the `Likelihood` instance based on the pseudo measurement
    - `full_measurement_likelihood`: the `MeasurementLikelihood` instance based
    on the original measurements
    - `sm_covariance`: the `SMCovariance` instance (that can be used to load
    and save the SM covariance matrix)
    - `exp_covariance`: the `MeasurementCovariance` instance (that can be used
    to load and save the experimental covariance matrix)

    Instances can be imported and exported from/to YAML using the `load`
    and `dump` methods.
    """

    _input_schema_dict = {
        'name': str,
        'par_obj': vol.All([dict], iio.coerce_par_obj),
        'fit_parameters': vol.Any(None, [str]),
        'nuisance_parameters': vol.Any(None, [str]),
        'observables':  vol.All([iio.coerce_observable_tuple], iio.list_deduplicate),
        'exclude_measurements': vol.Any(None, [str]),
        'include_measurements': vol.Any(None, [str]),
    }

    _output_schema_dict = {
        'name': str,
        'par_obj': iio.get_par_diff,
        'fit_parameters': vol.Any(iio.ensurelist, [str]),
        'nuisance_parameters': vol.Any(iio.ensurelist, [str]),
        'observables':  [iio.coerce_observable_dict],
        'exclude_measurements': vol.Any(iio.ensurelist, [str]),
        'include_measurements': vol.Any(iio.ensurelist, [str]),
    }

    def __init__(self, name,
                 par_obj=flavio.default_parameters,
                 fit_parameters=None,
                 nuisance_parameters='all',
                 observables=None,
                 exclude_measurements=None,
                 include_measurements=None,
                 ):
        self.par_obj = par_obj
        self.parameters_central = self.par_obj.get_central_all()
        self.fit_parameters = fit_parameters or []
        if nuisance_parameters == 'all':
            self.nuisance_parameters = self.par_obj.all_parameters
        else:
            self.nuisance_parameters = nuisance_parameters or []
        self.observables = observables
        self.exclude_measurements = exclude_measurements
        self.include_measurements = include_measurements
        self.full_measurement_likelihood = MeasurementLikelihood(
            self.observables,
            exclude_measurements=self.exclude_measurements,
            include_measurements=self.include_measurements)
        self.sm_covariance = SMCovariance(self.observables,
            vary_parameters=self.nuisance_parameters,
            par_obj=self.par_obj)
        self.exp_covariance = MeasurementCovariance(
            self.full_measurement_likelihood)
        self.pseudo_measurement = None
        self._likelihood = None
        NamedInstanceClass.__init__(self, name)

    def make_measurement(self, N=100, Nexp=5000, threads=1, force=False, force_exp=False):
        """Initialize the likelihood by producing a pseudo-measurement containing both
        experimental uncertainties as well as theory uncertainties stemming
        from nuisance parameters.

        Optional parameters:

        - `N`: number of random computations for the SM covariance (computing
          time is proportional to it; more means less random fluctuations.)
        - `Nexp`: number of random computations for the experimental covariance.
          This is much less expensive than the theory covariance, so a large
          number can be afforded (default: 5000).
        - `threads`: number of parallel threads for the SM
          covariance computation. Defaults to 1 (no parallelization).
        - `force`: if True, will recompute SM covariance even if it
          already has been computed. Defaults to False.
        - `force_exp`: if True, will recompute experimental central values and
          covariance even if they have already been computed. Defaults to False.
        """
        central_exp, cov_exp = self.exp_covariance.get(Nexp, force=force_exp)
        cov_sm = self.sm_covariance.get(N, force=force, threads=threads)
        covariance = cov_exp + cov_sm
        # add the Pseudo-measurement
        m = flavio.classes.Measurement('Pseudo-measurement for FastLikelihood instance: ' + self.name)
        if np.asarray(central_exp).ndim == 0 or len(central_exp) <= 1: # for a 1D (or 0D) array
            m.add_constraint(self.observables,
                    NormalDistribution(central_exp, np.sqrt(covariance)))
        else:
            m.add_constraint(self.observables,
                    MultivariateNormalDistribution(central_exp, covariance))
        self.pseudo_measurement = m
        self._likelihood = Likelihood(
            par_obj=self.par_obj,
            fit_parameters=self.fit_parameters,
            observables=self.observables,
            include_measurements=[m.name],  # only include our pseudo-meas.
            include_pseudo_measurements=True,  # force including the pseudo-meas.
            )

    @property
    def likelihood(self):
        if self._likelihood is None:
            raise ValueError("You need to call `make_measurement` first.")
        return self._likelihood

    def log_likelihood(self, par_dict, wc_obj, exclude_observables=None, delta=False):
        """Log-likelihood function.

        Parameters:
        - `par_dict`: a dictionary of parameter values
        - `wc_obj`: an instance of `WilsonCoefficients` or `wilson.Wilson`
        """
        return self.likelihood.log_likelihood_exp(par_dict, wc_obj, exclude_observables=exclude_observables, delta=delta)



r"""Functions for the branching ratio of the leptonic decay
$B_q\to\ell^+\ell^-\gamma$, where $q=d$ or $s$ and $\ell=e$, $\mu$ or
$\tau$. The branching ratio is taken from hep-ph/0410146"""

import flavio
from math import sqrt,pi,atanh
from flavio.physics.bdecays.common import meson_quark
from flavio.physics import ckm
from flavio.classes import AuxiliaryQuantity
from flavio.config import config
from flavio.physics.running import running
from flavio.physics.bdecays.wilsoncoefficients import get_wceff, wctot_dict
from flavio.classes import Observable, Prediction
import warnings

def _Re(z):
    return z.real
def _Im(z):
    return z.imag
def _Co(z):
    return complex(z).conjugate()

##################
# hep-ph/0410146:
#   - Uses the C7>0, C9<0, C10>0 convention
#   - has a global minus sign typo in dG12dsMN
#   - defines f_V instead of f_V^{.e.m} (see 1712.07926 and fix in the code)
#   - Defines an effective C9 (as for B->P/Vll), for the moment C9 is used,
#     which means that charmonium is not implemented!!
##################

# form factors
def get_ff(q2, par, B):
    return AuxiliaryQuantity['B->gamma form factor'].prediction(par_dict=par, wc_obj=None, q2=q2, B=B)


def prefactor(s, par, B, ff, lep, wc):
    GF = par['GF']
    scale = config['renormalization scale']['bllgamma']
    alphaem = running.get_alpha(par, scale)['alpha_e']
    bq = meson_quark[B]
    xi_t = ckm.xi('t',bq)(par)
    return GF**2/(2**10*pi**4)*abs(xi_t)**2*alphaem**3*par['m_'+B]**5

def getfft(s, par, B, ff, ff0, lep, wc):
    scale = config['renormalization scale']['bllgamma']
    mb = running.get_mb(par, scale, nf_out=5)
    bq = meson_quark[B]
    a1 = -wc['C1_'+bq]-wc['C2_'+bq]/3 #minus sign comes from WC sign convention
    fftv = ff['tv']+ff0['tv']+16/3*a1/wc['C7_'+bq]*par['f_'+B]/mb
    ffta = ff['ta']+ff0['ta']

    #Add light meson resonances
    resonances = {'Bs': ['phi'], 'B0': ['rho0', 'omega']}
    flavio.citations.register("Kozachuk:2017mdk")
    fVtofVemfactors = {'phi': -1/3, 'rho0': 1/sqrt(2), 'omega': 1/(3*sqrt(2))} # Given in Sec. 8.A.3 of 1712.07926
    #We use the general B->rho and B->omega parameters, hence the isotopic factors
    resgV = {('Bs','phi'): -par['Bs->phi BSZ a0_T1'],
             ('B0','rho0'): 1/sqrt(2)*par['B->rho BSZ a0_T1'],
             ('B0','omega'): -1/sqrt(2)*par['B->omega BSZ a0_T1'],}
    for V in resonances[B]:
        gV = resgV[(B, V)]
        fV = fVtofVemfactors[V]*par['f_'+V]
        fftv -= 2*fV*gV*par['m_'+B]**2*s/par['m_'+V]/(par['m_'+B]**2*s-par['m_'+V]**2+1j*par['m_'+V]/par['tau_'+V])
        ffta -= 2*fV*gV*par['m_'+B]**2*s/par['m_'+V]/(par['m_'+B]**2*s-par['m_'+V]**2+1j*par['m_'+V]/par['tau_'+V])

    return (fftv, ffta)

def getF1(s, par, B, ff, ff0, lep, wc):
    flavio.citations.register("Melikhov:2004mk")
    scale = config['renormalization scale']['bllgamma']
    mb = running.get_mb(par, scale, nf_out=5)
    mbh = mb/par['m_'+B]
    bq = meson_quark[B]
    label = bq+lep+lep
    fftv, ffta = getfft(s, par, B, ff, ff0, lep, wc)
    return (abs(wc['C9_'+label])**2 + abs(wc['C10_'+label])**2)*ff['v']**2 + 4*mbh**2/s**2*abs(wc['C7_'+bq]*fftv)**2 + 4*mbh/s*ff['v']*_Re(wc['C7_'+bq]*ffta*_Co(wc['C9_'+label]))

def getF2(s, par, B, ff, ff0, lep, wc):
    flavio.citations.register("Melikhov:2004mk")
    scale = config['renormalization scale']['bllgamma']
    mb = running.get_mb(par, scale, nf_out=5)
    mbh = mb/par['m_'+B]
    bq = meson_quark[B]
    label = bq+lep+lep
    fftv, ffta = getfft(s, par, B, ff, ff0, lep, wc)
    return (abs(wc['C9_'+label])**2 + abs(wc['C10_'+label])**2)*ff['a']**2 + 4*mbh**2/s**2*abs(wc['C7_'+bq]*ffta)**2 + 4*mbh/s*ff['a']*_Re(wc['C7_'+bq]*ffta*_Co(wc['C9_'+label]))


def B10(s, par, B, ff, ff0, lep, wc):
    flavio.citations.register("Melikhov:2004mk")
    flavio.citations.register("Guadagnoli:2016erb")
    F1 = getF1(s, par, B, ff, ff0, lep, wc)
    F2 = getF2(s, par, B, ff, ff0, lep, wc)
    mlh = par['m_'+lep]/par['m_'+B]
    label = meson_quark[B]+lep+lep
    return (s + 4*mlh**2)*(F1 + F2) - 8*mlh**2*abs(wc['C10_'+label])**2*(ff['v']**2+ff['a']**2)

# B11 vanishes for the BR estimation

def B12(s, par, B, ff, ff0, lep, wc):
    flavio.citations.register("Melikhov:2004mk")
    flavio.citations.register("Guadagnoli:2016erb")
    F1 = getF1(s, par, B, ff, ff0, lep, wc)
    F2 = getF2(s, par, B, ff, ff0, lep, wc)
    return s*(F1+F2)

def B120(s, par, B, ff, ff0, lep, wc):
    flavio.citations.register("Melikhov:2004mk")
    flavio.citations.register("Guadagnoli:2016erb")
    scale = config['renormalization scale']['bllgamma']
    mb = running.get_mb(par, scale, nf_out=5)
    mbh = mb/par['m_'+B]
    mlh = par['m_'+lep]/par['m_'+B]
    bq = meson_quark[B]
    label = bq+lep+lep
    fftv, ffta = getfft(s, par, B, ff, ff0, lep, wc)
    return -16*mlh**2*(1-s)*(ff['v']*_Re(wc['C9_'+label]*_Co(wc['C10_'+label])) + 2*mbh/s*_Re(_Co(wc['C10_'+label])*fftv*wc['C7_'+bq]))


def dG1dsMN(s, par, B, ff, ff0, lep, wc):
    flavio.citations.register("Melikhov:2004mk")
    mlh = par['m_'+lep]/par['m_'+B]
    pref = prefactor(s, par, B, ff, lep, wc)*(1-s)**3*sqrt(1-4*mlh**2/s)
    return pref*(B10(s, par, B, ff, ff0, lep, wc) + (s - 4*mlh**2)/(3*s)*B12(s, par, B, ff, ff0, lep, wc))

def dG2dsMN(s, par, B, ff, ff0, lep, wc):
    flavio.citations.register("Melikhov:2004mk")
    mlh = par['m_'+lep]/par['m_'+B]
    bq = meson_quark[B]
    label = bq+lep+lep
    pref = 16*prefactor(s, par, B, ff, lep, wc)*(par['f_'+B]/par['m_'+B]*abs(wc['C10_'+label])*mlh)**2/(1-s)
    return pref*(-8*sqrt(s-4*mlh**2)*sqrt(s)+4*atanh(sqrt(1-4*mlh**2/s))*(1+s-mlh**2*(1-s)**2))

def dG12dsMN(s, par, B, ff, ff0, lep, wc):
    flavio.citations.register("Melikhov:2004mk")
    mlh = par['m_'+lep]/par['m_'+B]
    pref = 4*prefactor(s, par, B, ff, lep, wc)*par['f_'+B]/par['m_'+B]*(1-s)
    return pref*atanh(sqrt(1-4*mlh**2/s))*B120(s, par, B, ff, ff0, lep, wc)

def dGdsMN(s, par, B, ff, ff0, lep, wc):
    return dG1dsMN(s, par, B, ff, ff0, lep, wc) + dG2dsMN(s, par, B, ff, ff0, lep, wc) + dG12dsMN(s, par, B, ff, ff0, lep, wc)


def bllg_dbrdq2(q2, wc_obj, par, B, lep):
    if q2 >= 8.7 and q2 < 14 and lep != 'tau':
        warnings.warn("The predictions in the region of narrow charmonium resonances are not meaningful")
    tauB = par['tau_'+B]
    ml = par['m_'+lep]
    mB = par['m_'+B]
    scale = config['renormalization scale']['bllgamma']
    ff = get_ff(q2, par, B)
    ff0 = get_ff(0, par, B)
    label = meson_quark[B]+lep+lep
    wc = wctot_dict(wc_obj, label, scale, par)
    if q2 <= 4*ml**2 or q2 > mB**2:
        return 0
    else:
        dBR = tauB * dGdsMN(q2/mB**2, par, B, ff, ff0, lep, wc)
        return dBR

def bllg_dbrdq2_func(B, lep):
    def fct(wc_obj, par, q2):
        return bllg_dbrdq2(q2, wc_obj, par, B, lep)
    return fct

def bllg_dbrdq2_int(q2min, q2max, wc_obj, par, B, lep, epsrel=0.005):
    def obs(q2):
        return bllg_dbrdq2(q2, wc_obj, par, B, lep)
    return flavio.math.integrate.nintegrate(obs, q2min, q2max, epsrel=epsrel)/(q2max-q2min)

def bllg_dbrdq2_int_func(B, lep):
    def fct(wc_obj, par, q2min, q2max):
        return bllg_dbrdq2_int(q2min, q2max, wc_obj, par, B, lep)
    return fct

_tex = {'e': 'e', 'mu': '\mu', 'tau': r'\tau'}
_tex_B = {'B0': r'\bar B^0', 'Bs': r'\bar B_s'}

_process_taxonomy = r'Process :: $b$ hadron decays :: FCNC decays :: $B\to \ell^+\ell^-\gamma$ :: $'

for l in ['e', 'mu']:
    for B in ['Bs', 'B0']:

        _process_tex =  _tex_B[B]+r"\to" +_tex[l]+r"^+"+_tex[l]+r"^-\gamma"


        # binned branching ratio
        _obs_name = "<dBR/dq2>("+B+"->"+l+l+"gamma)"
        _obs = Observable(name=_obs_name, arguments=['q2min', 'q2max'])
        _obs.set_description(r"Binned differential branching ratio of $" + _process_tex + r"$")
        _obs.tex = r"$\langle \frac{d\text{BR}}{dq^2} \rangle(" + _process_tex + r")$"
        _obs.add_taxonomy(_process_taxonomy + _process_tex + r"$")
        Prediction(_obs_name, bllg_dbrdq2_int_func(B, l))

        # differential branching ratio
        _obs_name = "dBR/dq2("+B+"->"+l+l+"gamma)"
        _obs = Observable(name=_obs_name, arguments=['q2'])
        _obs.set_description(r"Differential branching ratio of $" + _process_tex + r"$")
        _obs.tex = r"$\frac{d\text{BR}}{dq^2}(" + _process_tex + r")$"
        _obs.add_taxonomy(_process_taxonomy + _process_tex + r"$")
        Prediction(_obs_name, bllg_dbrdq2_func(B, l))



"""Functions to load and dump class instances from and to YAML dictionaries
or streams."""

import flavio
import voluptuous as vol
import yaml
from collections import OrderedDict


class YAMLLoadable(object):
    """Base class for objects that can be loaded and dumped from and to
    a dict or YAML stream."""

    # these class attributes should be overwritten by child classes
    _input_schema_dict = {}
    _output_schema_dict = {}

    @classmethod
    def input_schema(cls):
        return vol.Schema(cls._input_schema_dict, extra=vol.ALLOW_EXTRA)

    @classmethod
    def output_schema(cls):
        return vol.Schema(cls._output_schema_dict, extra=vol.REMOVE_EXTRA)

    @classmethod
    def load_dict(cls, d, **kwargs):
        """Instantiate an object from a YAML dictionary."""
        schema = cls.input_schema()
        return cls(**schema(d), **kwargs)

    @classmethod
    def load(cls, f, **kwargs):
        """Instantiate an object from a YAML string or stream."""
        d = flavio.io.yaml.load_include(f)
        return cls.load_dict(d, **kwargs)

    def get_yaml_dict(self):
        """Dump the object to a YAML dictionary."""
        d = self.__dict__.copy()
        schema = self.output_schema()
        d = schema(d)
        # remove NoneTypes and empty lists
        d = {k: v for k, v in d.items() if v is not None and v != []}
        return d

    def dump(self, stream=None, **kwargs):
        """Dump the objectto a YAML string or stream."""
        d = self.get_yaml_dict(**kwargs)
        return yaml.dump(d, stream=stream, **kwargs)


def coerce_observable_tuple(obs):
    """Force an arbitrary observable representation into the tuple representation."""
    return flavio.Observable.argument_format(obs, format='tuple')


def coerce_observable_dict(obs):
    """Force an arbitrary observable representation into the dict representation."""
    return flavio.Observable.argument_format(obs, format='dict')


def coerce_par_obj(par_obj_dict):
    """Coerce a dictionary of parameter constraints into a `ParameterConstraints`
    instance taking `flavio.default_parameters` as starting point"""
    par_obj = flavio.default_parameters.copy()
    return flavio.ParameterConstraints.from_yaml_dict(par_obj_dict,
                                                      instance=par_obj)


def ensurelist(v):
    """Coerce NoneType to empty list, wrap non-list in list."""
    if isinstance(v, list):
        return v
    elif v is None:
        return []
    else:
        raise ValueError("Unexpected form of list: {}".format(v))


def get_par_diff(par_obj):
    """Return a dictionary representation of a ParameterConstraints instance
    that only contains constraints that are not identical to ones in
    `default_parameters`."""
    dict_default = flavio.default_parameters.get_yaml_dict()
    dict_par = par_obj.get_yaml_dict()
    return [c for c in dict_par if c not in dict_default]


def list_deduplicate(lst):
    """Remove duplicate elements from a list but keep the order
    (keep the first occuring element of duplicates). List elements must be
    hashable."""
    return list(OrderedDict.fromkeys(lst))



import unittest
import flavio
from flavio.physics.units import s


par = flavio.default_parameters.get_central_all()


class TestBetaDecays(unittest.TestCase):
    def test_wceff(self):
        wc_obj = flavio.WilsonCoefficients()
        wceff = flavio.physics.betadecays.common.wc_eff(par, wc_obj, scale=1, nu='e')
        self.assertAlmostEqual(wceff['V'], 1, delta=0.05)
        self.assertAlmostEqual(wceff['A'], -1.27, delta=0.05)
        self.assertEqual(wceff['S'], 0)
        self.assertEqual(wceff['P'], 0)
        self.assertEqual(wceff['T'], 0)
        wc_obj = flavio.WilsonCoefficients()
        wc_obj.set_initial({'CVL_duenue': 1}, 1, 'WET-3', 'flavio')
        wceff = flavio.physics.betadecays.common.wc_eff(par, wc_obj, scale=1, nu='e')
        self.assertAlmostEqual(wceff['V'], 2, delta=0.05)
        self.assertAlmostEqual(wceff['A'], -1.27 * 2, delta=0.05)
        self.assertEqual(wceff['S'], 0)
        self.assertEqual(wceff['P'], 0)
        self.assertEqual(wceff['T'], 0)
        wc_obj.set_initial({'CVR_duenue': 1}, 1, 'WET-3', 'flavio')
        wceff = flavio.physics.betadecays.common.wc_eff(par, wc_obj, scale=1, nu='e')
        self.assertAlmostEqual(wceff['V'], 2, delta=0.05)
        self.assertAlmostEqual(wceff['A'], 0, delta=0.05)
        self.assertEqual(wceff['S'], 0)
        self.assertEqual(wceff['P'], 0)
        self.assertEqual(wceff['T'], 0)
        wc_obj.set_initial({'CT_duenue': 1}, 1, 'WET-3', 'flavio')
        wceff = flavio.physics.betadecays.common.wc_eff(par, wc_obj, scale=1, nu='e')
        self.assertAlmostEqual(wceff['V'], 1, delta=0.05)
        self.assertAlmostEqual(wceff['A'], -1.27, delta=0.05)
        self.assertEqual(wceff['S'], 0)
        self.assertEqual(wceff['P'], 0)
        self.assertAlmostEqual(wceff['T'], 4 * 1, delta=0.1)

    def test_ft(self):
        # compare to exp values in table 4 of 1803.08732
        wc_obj = flavio.WilsonCoefficients()
        Ft = flavio.physics.betadecays.ft.Ft_superallowed(par, wc_obj, '10C')
        self.assertAlmostEqual(Ft / s, 3078, delta=2 * 5)
        Ft = flavio.physics.betadecays.ft.Ft_superallowed(par, wc_obj, '26mAl')
        self.assertAlmostEqual(Ft / s, 3072.9, delta=3 * 1)
        Ft = flavio.physics.betadecays.ft.Ft_superallowed(par, wc_obj, '46V')
        self.assertAlmostEqual(Ft / s, 3074.1, delta=2 * 2)
        Ft = flavio.sm_prediction('Ft(38Ca)')
        self.assertAlmostEqual(Ft / s, 3076.4, delta=2 * 7.2)

    def test_taun(self):
        # compare to exp value in table 5 of 1803.08732
        tau_n = flavio.sm_prediction('tau_n', me_E=0.655)
        self.assertAlmostEqual(tau_n / s, 879.75, delta=39)

    def test_corrn(self):
        # compare to exp values in table 5 of 1803.08732
        self.assertAlmostEqual(flavio.sm_prediction('a_n'), -0.1034, delta=2 * 0.0037)
        self.assertAlmostEqual(flavio.sm_prediction('atilde_n', me_E=0.695), -0.1090, delta=0.01)
        self.assertAlmostEqual(flavio.sm_prediction('Atilde_n', me_E=0.569), -0.11869, delta=0.01)
        self.assertAlmostEqual(flavio.sm_prediction('Btilde_n', me_E=0.591), 0.9805, delta=3 * 0.003)
        self.assertAlmostEqual(flavio.sm_prediction('lambdaAB_n', me_E=0.581), -1.2686, delta=0.04)
        self.assertEqual(flavio.sm_prediction('D_n'), 0)
        self.assertEqual(flavio.sm_prediction('R_n'), 0)



"""Compare to parameters of https://arxiv.org/abs/1908.09398

Cf.
https://github.com/eos/eos/blob/8487191336d608fb9517979d84188420834d568a/eos/form-factors/mesonic-hqet_TEST.cc#L199-L358
https://github.com/eos/eos/blob/8487191336d608fb9517979d84188420834d568a/eos/form-factors/mesonic-hqet_TEST.cc#L1023-L1180
"""

import unittest
import flavio


par = flavio.default_parameters.copy()
par.set_constraint("CLN rho2_xi", 1.5)
par.set_constraint("CLN c_xi", +3.0 / 2)
par.set_constraint("CLN xi3", +6.0)
#par.set_constraint("xipppp(1)", -9.0)
par.set_constraint("chi_2(1)", +0.5)
par.set_constraint("chi_2p(1)", -1.0)
par.set_constraint("chi_2pp(1)", +2.0)
par.set_constraint("chi_3p(1)", -1.5)
par.set_constraint("chi_3pp(1)", +2.5)
par.set_constraint("eta(1)", +0.25)
par.set_constraint("etap(1)", -1.25)
par.set_constraint("etapp(1)", +1.75)
par.set_constraint("CLN l_1(1)", +0.5)
par.set_constraint("CLN l_2(1)", -2.0)
par.set_constraint("CLN l_3(1)", 0.0)
par.set_constraint("CLN l_4(1)", 0.0)
par.set_constraint("CLN l_5(1)", 0.0)
par.set_constraint("CLN l_6(1)", 0.0)
par.set_constraint("CLN lp_1(1)", 0.0)
par.set_constraint("CLN lp_2(1)", 0.0)
par.set_constraint("CLN lp_3(1)", 0.0)
par.set_constraint("CLN lp_4(1)", 0.0)
par.set_constraint("CLN lp_5(1)", 0.0)
par.set_constraint("CLN lp_6(1)", 0.0)
par.set_constraint("m_B0", 5.27942)
par.set_constraint("m_B+", 5.27942)
par.set_constraint("m_D+", 1.86723)
par.set_constraint("m_D0", 1.86723)
par.set_constraint("lambda_1", -0.30)
par.set_constraint("m_b", 4.2226)  # to reproduce mb1S=4.71 GeV

class TestCLNParam(unittest.TestCase):
    def test_btop(self):
        eps = 0.002
        def ff(q2):
            _par = par.get_central_all()
            return flavio.physics.bdecays.formfactors.b_p.cln.ff('B->D',
                q2, _par, scale=4.8,
                order_z=3, order_z_slp=1, order_z_sslp=1)

        self.assertAlmostEqual(ff(4)['f+'], -0.317099, delta=eps);
        self.assertAlmostEqual(ff(4)['f0'], -0.311925, delta=eps);
        self.assertAlmostEqual(ff(4)['fT'], -0.043808, delta=eps);

        self.assertAlmostEqual(ff(8)['f+'], +0.273187, delta=eps);
        self.assertAlmostEqual(ff(8)['f0'], +0.198352, delta=eps);
        self.assertAlmostEqual(ff(8)['fT'], +0.514150, delta=eps);

        self.assertAlmostEqual(ff(10)['f+'], +0.721643, delta=eps);
        self.assertAlmostEqual(ff(10)['f0'], +0.544512, delta=eps);
        self.assertAlmostEqual(ff(10)['fT'], +0.952830, delta=eps);


    def test_btov(self):
        eps=0.002

        def ff(q2):
            _par = par.get_central_all()
            return flavio.physics.bdecays.formfactors.b_v.cln.ff('B->D*', q2, _par, scale=4.8,
                order_z=3, order_z_slp=1, order_z_sslp=1)

        mB = par.get_central('m_B0')
        mV = par.get_central('m_D*+')

        def q2_w(w):
            return mB**2 + mV**2 - w * (2 * mB * mV)

        q2 = q2_w(w=1.0)
        h = {
            'A1': +0.899905,
            'A2': +0.036348,
            'A3': +0.552732,
            'V':  +1.217624,
            'T1': +0.961994,
            'T2': -0.198494,
            'T3': -0.665817
        }
        ff_eos = flavio.physics.bdecays.formfactors.b_v.cln.h_to_A(mB, mV, h, q2)
        ff_flav = ff(q2)
        for k in ff_flav:
            self.assertAlmostEqual(ff_eos[k], ff_flav[k], delta=eps, msg="Failed for w=1.0, {}".format(k))

        q2 = q2_w(w=1.4)
        h = {
            'A1': +0.684812,
            'A2': -0.075634,
            'A3': +0.702037,
            'V':  +0.896998,
            'T1': +0.686863,
            'T2': -0.104330,
            'T3': -0.195575
        }
        ff_eos = flavio.physics.bdecays.formfactors.b_v.cln.h_to_A(mB, mV, h, q2)
        ff_flav = ff(q2)
        for k in ff_flav:
            self.assertAlmostEqual(ff_eos[k], ff_flav[k], delta=eps, msg="Failed for w=1.4, {}".format(k))


import unittest
import flavio
import os

class TestYAML(unittest.TestCase):
    def test_load_include(self):
        test_file = os.path.join(os.path.dirname(__file__), '..', 'data', 'test', '1.yaml')
        with open(test_file, 'r') as f:
            d = flavio.io.yaml.load_include(f)
        self.assertDictEqual(d, {'key': {'key2': 0},
                                 'key3': [1, 2, 3, 4, 5, 6, 7, 8]})
    def test_load_include_error(self):
        test_str = "key: !include relative_path.yaml"
        with self.assertRaises(ValueError):
            d = flavio.io.yaml.load_include(test_str)
        test_str = "key: !include /absolute_path.yaml"
        with self.assertRaises(OSError):
            d = flavio.io.yaml.load_include(test_str)
    def test_include_absolute_path(self):
        test_file = os.path.join(os.path.dirname(__file__), '..', 'data', 'test', '2.yaml')
        test_str = "key: !include {}".format(test_file)
        d = flavio.io.yaml.load_include(test_str)
        self.assertDictEqual(d, {'key': {'key2': 0}})



r"""Functions for Higgs signal strengths."""

import flavio
from . import production
from . import decay
from . import width


prod_modes = {
    'ggF': {
        'desc': 'gluon fusion production',
        'tex': 'gg',
        'str': 'gg',
    },
    'hw': {
        'desc': '$W$ boson associated production',
        'tex': 'Wh',
        'str': 'Wh',
    },
    'hz': {
        'desc': '$Z$ boson associated production',
        'tex': 'Zh',
        'str': 'Zh',
    },
    'hv': {
        'desc': '$Z$ or $W$ boson associated production',
        'tex': 'Vh',
        'str': 'Vh',
    },
    'tth': {
        'desc': 'top pair associated production',
        'tex': r't\bar t h',
        'str': 'tth',
    },
    'vv_h': {
        'desc': 'weak boson fusion',
        'tex': r'\text{VBF}',
        'str': 'VBF',
    },
}

decay_modes = {
    'h_bb': {
        'tex': r'b\bar b',
        'str': 'bb',
        'tex_class': r'h\to ff',
    },
    'h_cc': {
        'tex': r'c\bar c',
        'str': 'cc',
        'tex_class': r'h\to ff',
    },
    'h_tautau': {
        'tex': r'\tau^+\tau^-',
        'str': 'tautau',
        'tex_class': r'h\to ff',
    },
    'h_mumu': {
        'tex': r'\mu^+\mu^-',
        'str': 'mumu',
        'tex_class': r'h\to ff',
    },
    'h_ww': {
        'tex': r'W^+W^-',
        'str': 'WW',
        'tex_class': r'h\to VV',
    },
    'h_zz': {
        'tex': r'ZZ',
        'str': 'ZZ',
        'tex_class': r'h\to VV',
    },
    'h_vv': {
        'tex': r'VV',
        'str': 'VV',
        'tex_class': r'h\to VV',
    },
    'h_zga': {
        'tex': r'Z\gamma',
        'str': 'Zgamma',
        'tex_class': r'h\to VV',
    },
    'h_gaga': {
        'tex': r'\gamma\gamma',
        'str': 'gammagamma',
        'tex_class': r'h\to VV',
    },
}


def higgs_signalstrength(wc_obj, par, name_prod, name_dec):
    scale = flavio.config['renormalization scale']['hdecays']
    C = wc_obj.get_wcxf(sector='all', scale=scale, par=par,
                        eft='SMEFT', basis='Warsaw')
    f_prod = getattr(production, name_prod)
    f_dec = getattr(decay, name_dec)
    return f_prod(C) * f_dec(C) / width.Gamma_h(par,  C)


def make_obs_higgs(name_prod, name_dec):
    d_dec = decay_modes[name_dec]
    d_prod = prod_modes[name_prod]
    process_tex = r"h \to {}".format(d_dec['tex'])
    process_taxonomy = r'Process :: Higgs production and decay :: $' + d_dec['tex_class'] + r'$ :: $' + process_tex + r"$"
    obs_name = "mu_{}(h->{})".format(d_prod['str'], d_dec['str'])
    obs = flavio.classes.Observable(obs_name)
    obs.set_description(r"Signal strength of $" + process_tex + r"$ from " + d_prod['desc'])
    obs.tex = r"$\mu_{" + d_prod['tex'] + r"}(" + process_tex + r")$"
    obs.add_taxonomy(process_taxonomy)

    def obs_fct(wc_obj, par):
        return higgs_signalstrength(wc_obj, par, name_prod, name_dec)

    flavio.classes.Prediction(obs_name, obs_fct)


for prod in prod_modes:
    for dec in decay_modes:
        make_obs_higgs(prod, dec)



r"""Functions for Higgs decay.

Most of the numerical coefficients have been obtained with MadGraph_aMC@NLO v2.6.5
along with SMEFTsim v2 in the alpha scheme.
"""

import flavio


def h_gg(C):
    r"""Higgs decay to two gluons normalized to the SM"""
    # obtained from an analytical one-loop calculation
    flavio.citations.register("Falkowski:2019hvp")
    np = (+39.29 * C['phiG']
          +0.121 * (C['phiBox'] - C['phiD'] / 4.)
          +0.061 * (C['ll_1221'] / 2 - C['phil3_22'] - C['phil3_11'])
          -0.129 * C['uphi_33']
          +0.123 * C['uphi_22']
          +0.239 * C['dphi_33']
          +0.025 * C['dphi_22']
          )
    return 1 + 1e6 * np.real

def h_gaga(C):
    r"""Higgs decay to two photons normalized to the SM"""
    # obtained from an analytical one-loop calculation
    flavio.citations.register("Falkowski:2019hvp")
    np = (-45.78 * C['phiB']
          -13.75 * C['phiW']
          +(25.09 - 0.242) * C['phiWB']  # tree - loop
          +0.121 * C['phiBox']
          -0.141 * C['phiD']
          +0.127 * (C['ll_1221'] / 2 - C['phil3_22'] - C['phil3_11'])
          +0.035 * C['uphi_33']
          -0.034 * C['uphi_22']
          -0.017 * C['dphi_33']
          -0.043 * C['ephi_33']
          -0.008 * C['ephi_22']
          )
    return 1 + 1e6 * np.real

def h_ww(C):
    r"""Higgs decay to two $W$ bosons normalized to the SM"""
    flavio.citations.register("Falkowski:2019hvp")
    np = (-0.092 * C['phiW']
          -0.386 * C['phiWB']
          -0.205 * C['phiD']
          +0.121 * C['phiBox']
          +0.144 * C['ll_1221']
          -0.167 * C['phil3_11']
          -0.167 * C['phil3_22']
          )
    return 1 + 1e6 * np.real

def h_zz(C):
    r"""Higgs decay to $Z$ bosons normalized to the SM"""
    flavio.citations.register("Falkowski:2019hvp")
    np = (+0.329 * C['phiB']
          -0.386 * C['phiW']
          +0.149 * C['phiWB']
          -0.039 * C['phiD']
          +0.118 * C['phiBox']
          +0.198 * (C['ll_1221'] / 2 - C['phil3_22'] - C['phil3_11'])
          )
    return 1 + 1e6 * np.real

def h_vv(C):
    r"""Higgs decay to $W$ or $Z$ bosons normalized to the SM"""
    br_ww = flavio.physics.higgs.width.BR_SM['WW']
    br_zz = flavio.physics.higgs.width.BR_SM['ZZ']
    d_ww = h_ww(C) - 1
    d_zz = h_zz(C) - 1
    return (br_ww * (1 + d_ww) + br_zz * (1 + d_zz)) / (br_ww + br_zz)

def h_zga(C):
    r"""Higgs decay to $Z\gamma$ normalized to the SM"""
    flavio.citations.register("Falkowski:2019hvp")
    np = (+14.89 * C['phiB']
          -14.89 * C['phiW']
          +9.377 * C['phiWB']
          )
    return 1 + 1e6 * np.real

def h_bb(C):
    r"""Higgs decay to two $b$ quarks normalized to the SM"""
    flavio.citations.register("Falkowski:2019hvp")
    np = (-0.03 * C['phiD']
          +0.121 * C['phiBox']
          +0.061 * (C['ll_1221'] / 2 - C['phil3_22'] - C['phil3_11'])
          -5.05 * C['dphi_33'])
    return 1 + 1e6 * np.real

def h_cc(C):
    r"""Higgs decay to two charm quarks normalized to the SM"""
    flavio.citations.register("Falkowski:2019hvp")
    np = (-0.03 * C['phiD']
          +0.121 * C['phiBox']
          +0.061 * (C['ll_1221'] / 2 - C['phil3_22'] - C['phil3_11'])
          -16.49 * C['uphi_22'])
    return 1 + 1e6 * np.real

def h_tautau(C):
    r"""Higgs decay to two taus normalized to the SM"""
    flavio.citations.register("Falkowski:2019hvp")
    np = (-0.03 * C['phiD']
          +0.121 * C['phiBox']
          +0.061 * (C['ll_1221'] / 2 - C['phil3_22'] - C['phil3_11'])
          -11.88 * C['ephi_33'])
    return 1 + 1e6 * np.real

def h_mumu(C):
    r"""Higgs decay to two muons normalized to the SM"""
    flavio.citations.register("Falkowski:2019hvp")
    np = (-0.03 * C['phiD']
          +0.121 * C['phiBox']
          +0.061 * (C['ll_1221'] / 2 - C['phil3_22'] - C['phil3_11'])
          -199.8 * C['ephi_22'])
    return 1 + 1e6 * np.real



"""Functions for effective semileptonic couplings."""

from flavio.physics.edms.common import proton_charges


def CS(wc, par, scale, Z, N):
    """Scalar electron-nucleon coupling."""
    me = proton_charges(par, scale)
    CSu = -(wc['CSRR_eeuu'].imag + wc['CSRL_eeuu'].imag)
    CSd = -(wc['CSRR_eedd'].imag + wc['CSRL_eedd'].imag)
    CS0 = me['gS_u+d'] * (CSu + CSd)
    CS1 = me['gS_u-d'] * (CSu - CSd)
    return CS0 + (Z - N) / (Z + N) * CS1
